from google.protobuf import message
import logging
import struct
from recall.proto import rpc_meta_pb2


def serialize_message(meta_info, msg):
    meta_info.msg_name = msg.DESCRIPTOR.full_name

    meta_buf = meta_info.SerializeToString()
    msg_buf = msg.SerializeToString()
    meta_buf_len = len(meta_buf)
    msg_buf_len = len(msg_buf)

    pb_buf_len = struct.pack('!III', meta_buf_len + msg_buf_len + 8,
                             meta_buf_len, msg_buf_len)
    msg_buf = 'PB' + pb_buf_len + meta_buf + msg_buf
    return msg_buf


def parse_message(buf, msg_cls):
    msg = msg_cls()
    try:
        msg.ParseFromString(buf)
        return msg
    except message.DecodeError as err:
        logging.warning('parsing msg failed: ' + str(err))
        return None


def parse_meta(buf, meta_info=None):
    if len(buf) < 8:
        return None

    (meta_len, pb_msg_len) = struct.unpack('!II', buf[:8])
    if len(buf) < 8 + meta_len + pb_msg_len:
        return None

    meta_msg_buf = buf[8:8 + meta_len]
    if meta_info is None:
        meta_info = rpc_meta_pb2.MetaInfo()

    try:
        meta_info.ParseFromString(meta_msg_buf)
        return meta_len, pb_msg_len, meta_info
    except message.DecodeError as err:
        logging.warning('parsing msg meta info failed: ' + str(err))
        return None




# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import service
from google.protobuf import service_reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)



DESCRIPTOR = descriptor.FileDescriptor(
  name='recall/proto/rpc_meta.proto',
  package='rpc',
  serialized_pb='\n\x1brecall/proto/rpc_meta.proto\x12\x03rpc\"k\n\x08MetaInfo\x12\x0f\n\x07\x66low_id\x18\x01 \x02(\x05\x12\x14\n\x0cservice_name\x18\x02 \x02(\t\x12\x13\n\x0bmethod_name\x18\x03 \x02(\t\x12\x10\n\x08msg_name\x18\x04 \x02(\t\x12\x11\n\thas_error\x18\x1e \x01(\x08\"B\n\rErrorResponse\x12 \n\x08\x65rr_code\x18\x01 \x02(\x0e\x32\x0e.rpc.ErrorCode\x12\x0f\n\x07\x65rr_msg\x18\x02 \x02(\t\"%\n\x10HeartBeatRequest\x12\x11\n\tmagic_num\x18\x01 \x02(\x05\"(\n\x11HeartBeatResponse\x12\x13\n\x0breturn_code\x18\x01 \x02(\x05*N\n\tErrorCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\x18\n\x14SERVER_SERVICE_ERROR\x10\x01\x12\x1a\n\x16SERVER_SERVICE_TIMEOUT\x10\x02\x32L\n\x0e\x42uiltinService\x12:\n\tHeartBeat\x12\x15.rpc.HeartBeatRequest\x1a\x16.rpc.HeartBeatResponseB\x03\x90\x01\x01')

_ERRORCODE = descriptor.EnumDescriptor(
  name='ErrorCode',
  full_name='rpc.ErrorCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SERVER_SERVICE_ERROR', index=1, number=1,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SERVER_SERVICE_TIMEOUT', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=294,
  serialized_end=372,
)


SUCCESS = 0
SERVER_SERVICE_ERROR = 1
SERVER_SERVICE_TIMEOUT = 2



_METAINFO = descriptor.Descriptor(
  name='MetaInfo',
  full_name='rpc.MetaInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='flow_id', full_name='rpc.MetaInfo.flow_id', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='service_name', full_name='rpc.MetaInfo.service_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='method_name', full_name='rpc.MetaInfo.method_name', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='msg_name', full_name='rpc.MetaInfo.msg_name', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='has_error', full_name='rpc.MetaInfo.has_error', index=4,
      number=30, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=36,
  serialized_end=143,
)


_ERRORRESPONSE = descriptor.Descriptor(
  name='ErrorResponse',
  full_name='rpc.ErrorResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='err_code', full_name='rpc.ErrorResponse.err_code', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='err_msg', full_name='rpc.ErrorResponse.err_msg', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=145,
  serialized_end=211,
)


_HEARTBEATREQUEST = descriptor.Descriptor(
  name='HeartBeatRequest',
  full_name='rpc.HeartBeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='magic_num', full_name='rpc.HeartBeatRequest.magic_num', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=213,
  serialized_end=250,
)


_HEARTBEATRESPONSE = descriptor.Descriptor(
  name='HeartBeatResponse',
  full_name='rpc.HeartBeatResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='return_code', full_name='rpc.HeartBeatResponse.return_code', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=252,
  serialized_end=292,
)

_ERRORRESPONSE.fields_by_name['err_code'].enum_type = _ERRORCODE
DESCRIPTOR.message_types_by_name['MetaInfo'] = _METAINFO
DESCRIPTOR.message_types_by_name['ErrorResponse'] = _ERRORRESPONSE
DESCRIPTOR.message_types_by_name['HeartBeatRequest'] = _HEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['HeartBeatResponse'] = _HEARTBEATRESPONSE

class MetaInfo(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _METAINFO
  
  # @@protoc_insertion_point(class_scope:rpc.MetaInfo)

class ErrorResponse(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _ERRORRESPONSE
  
  # @@protoc_insertion_point(class_scope:rpc.ErrorResponse)

class HeartBeatRequest(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _HEARTBEATREQUEST
  
  # @@protoc_insertion_point(class_scope:rpc.HeartBeatRequest)

class HeartBeatResponse(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _HEARTBEATRESPONSE
  
  # @@protoc_insertion_point(class_scope:rpc.HeartBeatResponse)


_BUILTINSERVICE = descriptor.ServiceDescriptor(
  name='BuiltinService',
  full_name='rpc.BuiltinService',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=374,
  serialized_end=450,
  methods=[
  descriptor.MethodDescriptor(
    name='HeartBeat',
    full_name='rpc.BuiltinService.HeartBeat',
    index=0,
    containing_service=None,
    input_type=_HEARTBEATREQUEST,
    output_type=_HEARTBEATRESPONSE,
    options=None,
  ),
])

class BuiltinService(service.Service):
  __metaclass__ = service_reflection.GeneratedServiceType
  DESCRIPTOR = _BUILTINSERVICE
class BuiltinService_Stub(BuiltinService):
  __metaclass__ = service_reflection.GeneratedServiceStubType
  DESCRIPTOR = _BUILTINSERVICE

# @@protoc_insertion_point(module_scope)



import gevent
import gevent.queue
import gevent.pool


class Pool(object):
    def __init__(self, pool_size=None):
        self._task_queue = gevent.queue.JoinableQueue()
        self._pool = gevent.pool.Pool(pool_size)
        if pool_size is None:
            pool_size = 100

        for _ in xrange(pool_size):
            self._pool.spawn(self.worker_func)

    def worker_func(self):
        while True:
            task = self._task_queue.get()
            if task is None:
                self._task_queue.task_done()
                break
            task()
            self._task_queue.task_done()

    def spawn(self, func, *args, **kwargs):
        task = lambda: func(*args, **kwargs)
        self._task_queue.put_nowait(task)

    def join(self):
        for _ in xrange(len(self._pool)):
            self._task_queue.put_nowait(None)
        self._task_queue.join()
        self._pool.join()

    def kill(self):
        self._pool.kill()


def kill_all_but_this(greenlets):
    this_greenlet = gevent.getcurrent()
    for gl in greenlets:
        if gl is not this_greenlet:
            gl.kill(block=False)




import google.protobuf


class RpcController(google.protobuf.service.RpcController):
    SUCCESS = 0
    SERVICE_TIMEOUT = 100
    WRONG_FLOW_ID_ERROR = 101
    WRONG_RSP_META_ERROR = 102
    WRONG_MSG_NAME_ERROR = 103
    SERVER_CLOSE_CONN_ERROR = 104
    CONN_FAILED_ERROR = 105

    def __init__(self, method_timeout=0, flow_id=None):
        self.err_code = RpcController.SUCCESS
        self.err_msg = None
        self.method_timeout = method_timeout
        self._flow_id = flow_id

    def Reset(self):
        self.err_code = RpcController.SUCCESS
        self.err_msg = None

    def Failed(self):
        return self.err_code != RpcController.SUCCESS

    def ErrorText(self):
        return self.err_msg

    def SetFailed(self, reason):
        self.err_code, self.err_msg = reason

    def SetFlowId(self, flow_id):
        assert isinstance(flow_id, int)
        self._flow_id = flow_id

    def GetFlowId(self):
        return self._flow_id




# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import service
from google.protobuf import service_reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)



DESCRIPTOR = descriptor.FileDescriptor(
  name='tests/test_proto/test.proto',
  package='test',
  serialized_pb='\n\x1btests/test_proto/test.proto\x12\x04test\"(\n\x0bTestRequest\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x0b\n\x03num\x18\x02 \x02(\x05\"0\n\x0cTestResponse\x12\x13\n\x0breturn_code\x18\x01 \x02(\x05\x12\x0b\n\x03msg\x18\x02 \x02(\t2B\n\x0bTestService\x12\x33\n\nTestMethod\x12\x11.test.TestRequest\x1a\x12.test.TestResponseB\x03\x90\x01\x01')




_TESTREQUEST = descriptor.Descriptor(
  name='TestRequest',
  full_name='test.TestRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='test.TestRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='num', full_name='test.TestRequest.num', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=37,
  serialized_end=77,
)


_TESTRESPONSE = descriptor.Descriptor(
  name='TestResponse',
  full_name='test.TestResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='return_code', full_name='test.TestResponse.return_code', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='msg', full_name='test.TestResponse.msg', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=79,
  serialized_end=127,
)

DESCRIPTOR.message_types_by_name['TestRequest'] = _TESTREQUEST
DESCRIPTOR.message_types_by_name['TestResponse'] = _TESTRESPONSE

class TestRequest(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _TESTREQUEST
  
  # @@protoc_insertion_point(class_scope:test.TestRequest)

class TestResponse(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _TESTRESPONSE
  
  # @@protoc_insertion_point(class_scope:test.TestResponse)


_TESTSERVICE = descriptor.ServiceDescriptor(
  name='TestService',
  full_name='test.TestService',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=129,
  serialized_end=195,
  methods=[
  descriptor.MethodDescriptor(
    name='TestMethod',
    full_name='test.TestService.TestMethod',
    index=0,
    containing_service=None,
    input_type=_TESTREQUEST,
    output_type=_TESTRESPONSE,
    options=None,
  ),
])

class TestService(service.Service):
  __metaclass__ = service_reflection.GeneratedServiceType
  DESCRIPTOR = _TESTSERVICE
class TestService_Stub(TestService):
  __metaclass__ = service_reflection.GeneratedServiceStubType
  DESCRIPTOR = _TESTSERVICE

# @@protoc_insertion_point(module_scope)



import random


class LoadBalancerBase(object):
    def get_connection_for_req(self, flow_id, req, conns):
        raise NotImplementedError


class SingleConnLoadBalancer(LoadBalancerBase):
    def get_connection_for_req(self, flow_id, req, conns):
        assert len(conns) == 1
        return conns[0]


class FixedLoadBalancer(LoadBalancerBase):
    def get_connection_for_req(self, flow_id, req, conns):
        return conns[flow_id % len(conns)]


class RandomLoadBalancer(LoadBalancerBase):
    def get_connection_for_req(self, flow_id, req, conns):
        return random.choice(conns)


class ReqNumLoadBalancer(LoadBalancerBase):
    def get_conn_req_num(self, conn):
        return conn.get_pending_send_task_num()

    def get_connection_for_req(self, flow_id, req, conns):
        return min(conns, key=self.get_conn_req_num)


class DelayLoadBalancer(LoadBalancerBase):
    def __init__(self, is_random=True):
        self._call_times = 0
        self._is_random = is_random

    def get_conn_avg_delay(self, conn):
        return conn.get_avg_delay_per_min()

    def get_delay_index(self, conns):
        delay_idxs = []
        sum_delay = 0.0
        for conn in conns:
            delay = self.get_conn_avg_delay(conn)
            delay_index = 1.0 / delay if delay > 0.0 else 0.01
            if delay_index < 0.01:
                delay_index = 0.01
            sum_delay += delay_index
            delay_idxs.append(delay_index)

        return sum_delay, delay_idxs

    def get_connection_for_req(self, flow_id, req, conns):
        sum_delay, delay_idxs = self.get_delay_index(conns)
        rand_value = random.uniform(0, sum_delay)
        for i, delay in enumerate(delay_idxs):
            rand_value -= delay
            if rand_value <= 0.0:
                return conns[i]

        # should not reach here
        return random.choice(conns)




import struct
import google.protobuf.service
from socket import error as soc_error
import logging
import time
import heapq

import gevent
import gevent.socket
import gevent.queue
import gevent.event

from recall import loadbalance
from recall.codec import serialize_message, parse_message, parse_meta
from recall.controller import RpcController
from recall.proto import rpc_meta_pb2
from recall.util import Pool, kill_all_but_this


class TcpConnectionStat(object):
    def __init__(self):
        self.total_req_num_per_min = 0
        self.total_rsp_num_per_min = 0
        self.total_delay_s_per_min = 0
        self.max_delay_s_per_min = -1
        self.min_delay_s_per_min = -1

    def add_req_stat(self, count=1):
        self.total_req_num_per_min += count

    def add_rsp_stat(self, count, delay_s):
        self.total_rsp_num_per_min += count
        self.total_delay_s_per_min = delay_s

        if delay_s > self.max_delay_s_per_min:
            self.max_delay_s_per_min = delay_s

        if self.min_delay_s_per_min < 0 or delay_s < self.min_delay_s_per_min:
            self.min_delay_s_per_min = delay_s

    @property
    def avg_delay_s_per_min(self):
        if self.total_rsp_num_per_min <= 0:
            return -1
        else:
            return self.total_delay_s_per_min / self.total_rsp_num_per_min

    def reset_stat_per_min(self):
        self.total_req_num_per_min = 0
        self.total_rsp_num_per_min = 0
        self.total_delay_s_per_min = 0
        self.min_delay_s_per_min = -1
        self.max_delay_s_per_min = -1


class TcpConnection(object):

    socket_cls = gevent.socket.socket

    class Exception(Exception):
        def __init__(self, err_code, err_msg):
            self.err_code, self.err_msg = err_code, err_msg

        def __str__(self):
            return "<TcpConnectionError (%d, '%s')>" % \
                   (self.err_code, self.err_msg)

    class RequestData(object):
        def __init__(self, is_async, *call_args):
            self.begin_time = 0
            self.is_async = is_async
            self.call_args = call_args
            self.rpc_controller = call_args[2]
            self.req_msg = call_args[3]

    # state enumeration definition
    CLOSED = 0
    CONNECTED = 1
    UNHEALTHY = 2

    def __init__(self, addr, state_evt_listener, get_flow_func,
                 spawn=gevent.spawn, heartbeat_interval=30):
        self._state = TcpConnection.CLOSED
        self._state_evt_listener = state_evt_listener
        self._get_flow_func = get_flow_func
        self._addr = addr
        self._spawn = spawn
        self._heartbeat_interval = heartbeat_interval
        self._heartbeat_timeout = 3
        self._socket = None
        self._is_closed = True

        self._send_task_queue = gevent.queue.Queue()
        self._recv_infos = {}
        self._timeout_queue = []
        self._workers = []
        self._stat = TcpConnectionStat()

        self._recv_meta_info = rpc_meta_pb2.MetaInfo()

    def __del__(self):
        self.close()

    def __str__(self):
        return '<TcpConnection %s>' % str(self._addr)

    def _spawn_workers(self):
        self._workers = [gevent.spawn(self.send_loop), gevent.spawn(self.recv_loop),
                         gevent.spawn(self.timeout_loop), gevent.spawn(self.heartbeat_loop)]

    def get_addr(self):
        return self._addr

    def is_connected(self):
        return self._socket is not None

    # client should call connect after getting TcpConnection
    def connect(self):
        try:
            assert self._socket is None
            self._socket = TcpConnection.socket_cls()
            self._socket.connect(self._addr)
            self._socket.setsockopt(gevent.socket.SOL_TCP, gevent.socket.TCP_NODELAY, 1)
            self._socket.setsockopt(gevent.socket.IPPROTO_TCP, gevent.socket.TCP_NODELAY, 1)
            self.change_state(TcpConnection.CONNECTED)
            self._is_closed = False

            self._spawn_workers()
        except gevent.socket.error, e:
            self._socket = None
            raise TcpConnection.Exception(e.errno, e.strerror)

    def close(self):
        if self._is_closed:
            logging.warning('conn is closed')
            return

        self._socket.close()
        self._is_closed = True
        self.change_state(TcpConnection.CLOSED)

        # clear all data
        while not self._send_task_queue.empty():
            try:
                self._send_task_queue.get_nowait()
            except gevent.queue.Empty:
                pass

        self._timeout_queue = []

        # check if there is any request has not been processed.
        for v in self._recv_infos.itervalues():
            expected_meta, rpc_controller, response_class, done, req_data = v
            rpc_controller.SetFailed((RpcController.SERVER_CLOSE_CONN_ERROR,
                                      'channel has been closed prematurely'))
            self._finish_rpc(done, rpc_controller, None, req_data.is_async)

        self._recv_infos.clear()

        # socket has to be set to None here, because in _finish_rpc
        # the control flow may be switched to another greenlet.
        # And in TcpChannel, it may reconnect the connection.
        # If we don't set it at last, it may cause problem.
        self._socket = None

        if len(self._workers) > 0:
            kill_all_but_this(self._workers)

    def change_state(self, state):
        if self._state != state:
            old_state = self._state
            self._state = state
            self._state_evt_listener(self, old_state, state)

    # when change to unhealthy state,
    # TcpChannel will call this functions to get all
    # unsent task to dispatch to other good connections.
    def get_pending_send_tasks(self):
        unsent_tasks = []
        while not self._send_task_queue.empty():
            try:
                task = self._send_task_queue.get_nowait()
                req_data = task.req_data
                # all heartbeat requests and request targetting only to this
                # connection should not be sent in other connections
                if not isinstance(req_data.req_msg, rpc_meta_pb2.HeartBeatRequest) and \
                    req_data.rpc_controller.GetFlowId() is None:
                    unsent_tasks.append((req_data.is_async, req_data.call_args))
            except gevent.queue.Empty:
                pass

        logging.debug('unsent task num: %d' % len(unsent_tasks))
        return unsent_tasks

    def get_stat(self):
        return self._stat

    def get_avg_delay_per_min(self):
        return self.get_stat().avg_delay_s_per_min

    def get_pending_send_task_num(self):
        return self._send_task_queue.qsize()

    def add_send_task(self, flow_id, method_descriptor, rpc_controller,
                      request, response_class, done, req_data):
        send_task = lambda: self.send_req(flow_id, method_descriptor, rpc_controller,
                                          request, response_class, done, req_data)
        send_task.req_data = req_data

        self._do_add_send_task(send_task)

    def _do_add_send_task(self, send_task):
        self._send_task_queue.put_nowait(send_task)

    def _finish_rpc(self, done, controller, rsp, is_async):
        if is_async:
            self._spawn(done, controller, rsp)
        else:
            done(controller, rsp)

    def timeout_loop(self):
        while True:
            gevent.sleep(1)
            if len(self._timeout_queue) > 0:
                now = time.time()
                while len(self._timeout_queue) > 0:
                    deadline, flow_id = self._timeout_queue[0]  # the smallest element
                    if deadline > now:
                        break
                    else:
                        heapq.heappop(self._timeout_queue)
                        recv_info = self._recv_infos.get(flow_id)
                        if recv_info is None:
                            continue

                        meta_info, rpc_controller, response_class, done, req_data = \
                            recv_info
                        timeout_time = time.time()
                        self._stat.add_rsp_stat(1, timeout_time - req_data.begin_time)
                        err_msg = 'service timeout'
                        rpc_controller.SetFailed((RpcController.SERVICE_TIMEOUT, err_msg))
                        logging.warning(err_msg)
                        del self._recv_infos[flow_id]
                        self._finish_rpc(done, rpc_controller, None, req_data.is_async)

    def heartbeat_loop(self):
        service_descriptor = rpc_meta_pb2.BuiltinService_Stub.GetDescriptor()
        method_descriptor = service_descriptor.FindMethodByName('HeartBeat')
        request = rpc_meta_pb2.HeartBeatRequest(magic_num=4231)
        response_class = rpc_meta_pb2.HeartBeatResponse
        hb_fail_count = 0

        while True:
            gevent.sleep(self._heartbeat_interval)

            res = gevent.event.AsyncResult()
            flow_id = self._get_flow_func()
            rpc_controller = RpcController(method_timeout=self._heartbeat_timeout)
            done = lambda _, rsp: res.set(rsp)
            req_data = self.RequestData(False, flow_id, method_descriptor, rpc_controller,
                                        request, response_class, done)
            self.add_send_task(flow_id, method_descriptor, rpc_controller,
                               request, response_class, done,
                               req_data)
            hb_rsp = res.get()
            if hb_rsp is None:
                logging.warning('connection %s heartbeat timeout' % str(self._addr))
                hb_fail_count += 1
                if hb_fail_count >= 3:
                    logging.warning('connection %s become unhealthy' % str(self._addr))
                    self.change_state(TcpConnection.UNHEALTHY)
            else:
                hb_fail_count = 0
                self.change_state(TcpConnection.CONNECTED)

    def recv_loop(self):
        try:
            while True:
                if not self.recv_rsp():
                    break
        except Exception, e:
            logging.warning('recv_loop failed and exit: ' + str(e))
        finally:
            self.close()

    def send_loop(self):
        while True:
            send_task = self._send_task_queue.get()
            if not send_task():
                break

    def send_req(self, flow_id, method_descriptor, rpc_controller,
                 request, response_class, done, req_data):
        service_descriptor = method_descriptor.containing_service
        meta_info = rpc_meta_pb2.MetaInfo()
        meta_info.flow_id = flow_id
        meta_info.service_name = service_descriptor.full_name
        meta_info.method_name = method_descriptor.name
        serialized_req = serialize_message(meta_info, request)
        sent_bytes = 0
        try:
            while sent_bytes < len(serialized_req):
                sent_bytes += self._socket.send(serialized_req[sent_bytes:])
        except soc_error as e:
            logging.warning('socket error: ' + str(e))
            rpc_controller.SetFailed((RpcController.SERVER_CLOSE_CONN_ERROR,
                                      'connection has sending error'))
            self._finish_rpc(done, rpc_controller, None, req_data.is_async)

            self.close()
            return False

        req_data.begin_time = time.time()
        self._recv_infos[flow_id] = (meta_info, rpc_controller, response_class, done, req_data)
        self._stat.add_req_stat()
        if rpc_controller.method_timeout > 0:
            heapq.heappush(self._timeout_queue,
                           (req_data.begin_time + rpc_controller.method_timeout, flow_id))

        return True

    def _recv(self, expected_size):
        recv_buf = ''
        while len(recv_buf) < expected_size:
            try:
                buf = self._socket.recv(expected_size - len(recv_buf))
                if len(buf) == 0:
                    break

                recv_buf += buf
            except Exception, e:
                logging.warning('recv failed: ' + str(e))
                raise

        return recv_buf

    _expected_size = 2 + struct.calcsize("!I")
    _error_msg_name = rpc_meta_pb2.ErrorResponse.DESCRIPTOR.full_name

    # return True when the connection is not closed
    # return False when it's closed
    def recv_rsp(self):
        expected_size = TcpConnection._expected_size
        pb_buf = self._recv(expected_size)
        if len(pb_buf) == 0:
            logging.info('socket has been closed')
            return False
        if len(pb_buf) < expected_size:
            logging.warning('rsp buffer broken')
            return True
        if pb_buf[:2] != 'PB':
            logging.warning('rsp buffer not begin with PB')
            return True

        buf_size = struct.unpack("!I", pb_buf[2:])[0]
        pb_buf = self._recv(buf_size)
        if len(pb_buf) == 0:
            logging.info('socket has been closed')
            return False

        result = parse_meta(pb_buf, self._recv_meta_info)
        if result is None:
            logging.warning('pb decode error, skip this message')
            return True

        meta_len, pb_msg_len, meta_info = result

        if meta_info.flow_id in self._recv_infos:
            expected_meta, rpc_controller, response_class, done, req_data = \
                self._recv_infos[meta_info.flow_id]

            try:
                if meta_info.flow_id != expected_meta.flow_id:
                    err_msg = 'rsp flow id not match: %d %d' % (expected_meta.flow_id,
                                                                meta_info.flow_id)
                    raise TcpConnection.Exception(RpcController.WRONG_FLOW_ID_ERROR, err_msg)
                elif meta_info.service_name != expected_meta.service_name or \
                                meta_info.method_name != expected_meta.method_name:
                    err_msg = 'rsp meta not match'
                    raise TcpConnection.Exception(RpcController.WRONG_RSP_META_ERROR, err_msg)
                elif meta_info.HasField('has_error'):
                    if meta_info.msg_name != TcpConnection._error_msg_name:
                        err_msg = 'rsp meta has error, but with wrong msg name'
                        raise TcpConnection.Exception(RpcController.WRONG_MSG_NAME_ERROR, err_msg)
                    else:
                        response_class = rpc_meta_pb2.ErrorResponse
                elif meta_info.msg_name != response_class.DESCRIPTOR.full_name:
                    err_msg = 'wrong response class'
                    raise TcpConnection.Exception(RpcController.WRONG_MSG_NAME_ERROR, err_msg)

                rsp_time = time.time()
                self._stat.add_rsp_stat(1, rsp_time - req_data.begin_time)
                rsp = parse_message(pb_buf[8 + meta_len:8 + meta_len + pb_msg_len],
                                     response_class)

                if meta_info.HasField('has_error'):
                    raise TcpConnection.Exception(rsp.err_code, rsp.err_msg)

                del self._recv_infos[meta_info.flow_id]
                self._finish_rpc(done, rpc_controller, rsp, req_data.is_async)
                return True
            except TcpConnection.Exception as e:
                rpc_controller.SetFailed((e.err_code, e.err_msg))
                logging.warning(e.err_msg)
                del self._recv_infos[meta_info.flow_id]
                self._finish_rpc(done, rpc_controller, None, req_data.is_async)
                return True
        else:
            logging.warning('flow id not found: %d' % meta_info.flow_id)
            return True


class TcpChannel(google.protobuf.service.RpcChannel):

    conn_cls = TcpConnection
    _default_load_balancer = loadbalance.DelayLoadBalancer()
    _fixed_load_balancer = loadbalance.FixedLoadBalancer()

    def __init__(self, addr, load_balancer=None,
                 spawn=gevent.spawn, connect_interval=30):
        google.protobuf.service.RpcChannel.__init__(self)
        self._flow_id = 0
        self._addr = addr
        self._user_load_balancer = load_balancer
        self._balancer = None
        self._spawn = spawn

        self._all_connections = []
        self._good_connections = []
        self._bad_connections = []
        # TODO: handle addr format error
        for ip_port in self.resolve_addr(addr):
            conn = TcpChannel.conn_cls(ip_port, self.on_conn_state_changed,
                                       self._get_flow_id, spawn=spawn)
            # because these connections is not connected,
            # put them to bad connections
            self._all_connections.append(conn)

            # no need to add conn because it will
            # add itself in state_changed function.

        self._connected_event = gevent.event.Event()
        self._last_connect_time = time.time()
        self._connect_interval = connect_interval

    def __del__(self):
        self.close()

    def _get_flow_id(self):
        flow_id = self._flow_id
        self._flow_id += 1
        return flow_id

    # after close is called, the channel cannot be used
    def close(self):
        for conn in self._all_connections:
            if conn.is_connected():
                conn.close()

        self._good_connections = []
        self._bad_connections = []
        self._all_connections = []

    def connect(self):
        self._last_connect_time = time.time()
        if not self.is_connected():
            self._do_connect()

    def _do_connect(self):
        for conn in self._all_connections:
            try:
                if not conn.is_connected():
                    conn.connect()
            except TcpConnection.Exception, e:
                logging.warning('connection %s failed: %s' % (conn, str(e)))
                # just ignore it and continue connect next one

    def is_connected(self):
        return len(self._good_connections) > 0
    
    def wait_until_connected(self):
        while len(self._good_connections) == 0:
            self._connected_event.wait()

        self._connected_event.clear()

    def on_conn_state_changed(self, conn, old_state, new_state):
        try:  # TODO: this try block should be removed in production
            if new_state == TcpConnection.CONNECTED:
                if old_state == TcpConnection.UNHEALTHY:
                    self._bad_connections.remove(conn)
                self._good_connections.append(conn)
            elif new_state == TcpConnection.UNHEALTHY:
                assert old_state == TcpConnection.CONNECTED
                self._good_connections.remove(conn)
                self._bad_connections.append(conn)

                if len(self._good_connections) == 0:
                    logging.warning('All connections in channel is unhealthy.')
                else:
                    unsent_tasks = conn.get_pending_send_tasks()
                    for is_async, call_args in unsent_tasks:
                        (flow_id, method_descriptor,
                         rpc_controller, request, response_class,
                         done) = call_args
                        conn = self._load_balance(flow_id, rpc_controller, request)
                        self._call_method_on_conn(conn, flow_id, method_descriptor,
                                                  rpc_controller, request, response_class,
                                                  done, is_async)

            else:  # TcpConnection.CLOSED
                if old_state == TcpConnection.CONNECTED:
                    self._good_connections.remove(conn)
                else:  # TcpConnection.UNHEALTHY
                    self._bad_connections.remove(conn)

            if len(self._good_connections) == 0:
                return
            else:
                self._connected_event.set()

            # choose load balancer
            if len(self._good_connections) == 1:
                self._balancer = loadbalance.SingleConnLoadBalancer()
            elif self._user_load_balancer is not None:
                self._balancer = self._user_load_balancer
            else:
                self._balancer = TcpChannel._default_load_balancer
            logging.info('use %s as load balancer' % self._balancer.__class__.__name__)
        except:
            assert False
            raise

    def resolve_addr(self, addr):
        if isinstance(addr, (list, set, tuple)):
            ip_ports = []
            for ad in addr:
                ads = self.resolve_addr(ad)
                if ads is None:
                    return None
                ip_ports += ads
            return ip_ports
        elif isinstance(addr, str):
            sep_index = addr.find('/')
            if sep_index == -1:  # cannot find '/', so treat it as ip port
                try:
                    ip, port = addr.split(':')
                except ValueError:
                    return None

                port = int(port)
                return [(ip, port)]
            else:
                addr_type = addr[:sep_index]
                addr_str = addr[sep_index + 1:]
                if addr_type == 'ip':
                    ip, port = addr_str.split(':')
                    port = int(port)
                    return [(ip, port)]
                elif addr_type == 'zk':
                    raise NotImplementedError
                else:
                    raise NotImplementedError

    def _call_method_on_conn(self, conn, flow_id, method_descriptor,
                             rpc_controller, request, response_class,
                             done, is_async=None):
        if is_async is not None:
            req_data = TcpConnection.RequestData(is_async, flow_id,
                                method_descriptor, rpc_controller,
                                request, response_class, done)
            conn.add_send_task(flow_id, method_descriptor, rpc_controller,
                               request, response_class, done, req_data)
            return None
        elif done is None:
            req_data = TcpConnection.RequestData(False, flow_id,
                                method_descriptor, rpc_controller,
                                request, response_class, done)
            res = gevent.event.AsyncResult()
            done = lambda _, rsp: res.set(rsp)
            conn.add_send_task(flow_id, method_descriptor, rpc_controller,
                               request, response_class, done, req_data)
            return res.get()
        else:
            req_data = TcpConnection.RequestData(True, flow_id,
                                method_descriptor, rpc_controller,
                                request, response_class, done)
            conn.add_send_task(flow_id, method_descriptor, rpc_controller,
                               request, response_class, done, req_data)
            return None

    def _load_balance(self, flow_id, rpc_controller, request):
        conn_flow_id = flow_id
        user_flow_id = rpc_controller.GetFlowId()
        balancer = self._balancer
        if user_flow_id is not None:
            conn_flow_id = user_flow_id
            balancer = TcpChannel._fixed_load_balancer
        conn = balancer.get_connection_for_req(conn_flow_id, request, self._good_connections)
        return conn

    # when done is None, it means the method call is synchronous
    # when it's not None, it means the call is asynchronous
    def CallMethod(self, method_descriptor, rpc_controller,
                   request, response_class, done):
        if len(self._all_connections) > \
                len(self._good_connections) + len(self._bad_connections):
            now = time.time()
            if now - self._last_connect_time >= self._connect_interval:
                # connect async
                logging.info('try reconnect')
                self._last_connect_time = now
                self._spawn(self._do_connect)

        if len(self._good_connections) == 0:
            rpc_controller.SetFailed((RpcController.CONN_FAILED_ERROR,
                                      'No connection is valid'))
            if done is not None:
                self._spawn(done, rpc_controller, None)
            return None

        flow_id = self._get_flow_id()
        conn = self._load_balance(flow_id, rpc_controller, request)

        return self._call_method_on_conn(conn, flow_id, method_descriptor,
                                         rpc_controller, request,
                                         response_class, done)


class RpcClient(object):

    tcp_channel_class = TcpChannel

    def __init__(self, pool_size=1000):
        self._channels = {}
        self._pool = Pool(pool_size)

    def __del__(self):
        self.close()

    def close(self):
        for channel in self._channels.itervalues():
            channel.close()
        self._channels.clear()

        self._pool.join()

    def get_tcp_channel(self, addr, is_wait_connected=True):
        if isinstance(addr, list):
            addr = tuple(addr)
        if addr not in self._channels:
            # if TcpConnection connects for a long time, and
            # other client use get_tcp_channel may cause problem.
            channel = RpcClient.tcp_channel_class(addr, spawn=self._pool.spawn)
            self._channels[addr] = channel

            # connect asynchronously
            self._pool.spawn(channel.connect)
        else:
            channel = self._channels[addr]
            if not channel.is_connected():
                self._pool.spawn(channel.connect)

        if is_wait_connected:
            channel.wait_until_connected()

        return channel





from setuptools import setup, find_packages
from distutils.spawn import find_executable
import sys
import os
import subprocess

version = '0.2.1'

# Find the Protocol Compiler.
if 'PROTOC' in os.environ and os.path.exists(os.environ['PROTOC']):
    protoc = os.environ['PROTOC']
elif os.path.exists("../src/protoc"):
    protoc = "../src/protoc"
elif os.path.exists("../src/protoc.exe"):
    protoc = "../src/protoc.exe"
elif os.path.exists("../vsprojects/Debug/protoc.exe"):
    protoc = "../vsprojects/Debug/protoc.exe"
elif os.path.exists("../vsprojects/Release/protoc.exe"):
    protoc = "../vsprojects/Release/protoc.exe"
else:
    protoc = find_executable("protoc")


def generate_proto(source):
    """Invokes the Protocol Compiler to generate a _pb2.py from the given
    .proto file.  Does nothing if the output already exists and is newer than
    the input."""

    output = source.replace(".proto", "_pb2.py").replace("../src/", "")

    if (not os.path.exists(output) or
        (os.path.exists(source) and
         os.path.getmtime(source) > os.path.getmtime(output))):
        print ("Generating %s..." % output)

        if not os.path.exists(source):
            sys.stderr.write("Can't find required file: %s\n" % source)
            sys.exit(-1)

        if protoc == None:
            sys.stderr.write(
               "protoc is not installed nor found in ../src.  Please compile it "
               "or install the binary package.\n")
            sys.exit(-1)

        protoc_command = [ protoc, "-I.", "--python_out=.", source ]
        if subprocess.call(protoc_command) != 0:
            sys.exit(-1)


if __name__ == '__main__':
    # generate necessary proto files
    generate_proto('recall/proto/rpc_meta.proto')
    generate_proto('tests/test_proto/test.proto')

    setup(name='recall',
          version=version,
          description="Python High performance RPC framework based on protobuf",
          classifiers=[
              'Development Status :: 3 - Alpha',
              'Environment :: Web Environment',
              'Intended Audience :: Developers',
              'License :: OSI Approved :: MIT License',
              'Operating System :: POSIX',
              'Programming Language :: Python',
              'Programming Language :: Python :: 2.7',
              'Topic :: Internet'
	  ],
          keywords='rpc gevent',
          author='Yaolong Huang',
          author_email='airekans@gmail.com',
          url='https://github.com/airekans/recall',
          license='MIT',
          packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
          include_package_data=True,
          zip_safe=False,
          install_requires=[
              'gevent',
              'protobuf>=2.3'
          ],
          entry_points="""
          # -*- Entry points: -*-
          """,
      )



from _socket import error as soc_error
import logging
import struct
import traceback
import gevent
import gevent.server
import gevent.socket
from gevent import Timeout
from recall.codec import serialize_message, parse_message, parse_meta
from recall.controller import RpcController
from recall.proto import rpc_meta_pb2
from recall.util import Pool


class RpcServerStat(object):

    class MethodStat(object):
        def __init__(self):
            self.call_num_per_min = 0
            self.total_call_num = 0

        def reset_stat_per_min(self):
            self.call_num_per_min = 0

    class ServiceStat(object):
        def __init__(self):
            self.call_num_per_min = 0
            self.total_call_num = 0
            self.method_stats = {}

        def add_method_stat(self, method, count):
            try:
                method_stat = self.method_stats[method]
            except KeyError:
                method_stat = RpcServerStat.MethodStat()
                self.method_stats[method] = method_stat

            method_stat.call_num_per_min += count
            method_stat.total_call_num += count

        def reset_stat_per_min(self):
            self.call_num_per_min = 0
            for method_stat in self.method_stats.itervalues():
                method_stat.reset_stat_per_min()

    def __init__(self):
        self.service_stats = {}

    def add_method_stat(self, service, method, count):
        try:
            service_stat = self.service_stats[service]
        except KeyError:
            service_stat = RpcServerStat.ServiceStat()
            self.service_stats[service] = service_stat

        service_stat.call_num_per_min += count
        service_stat.total_call_num += count
        service_stat.add_method_stat(method, count)

    def print_stat(self):
        print 'name\t| total_count\t| count/60s'

        for service_name, service_stat in self.service_stats.iteritems():
            print '%s\t| %d\t| %d' % (service_name, service_stat.total_call_num,
                                      service_stat.call_num_per_min)
            for method_name, method_stat in service_stat.method_stats.iteritems():
                print '   %s\t| %d\t| %d' % (method_name, method_stat.total_call_num,
                                             method_stat.call_num_per_min)

        print ''

    def reset_stat_per_min(self):
        for service_stat in self.service_stats.itervalues():
            service_stat.reset_stat_per_min()


class BuiltinServiceImpl(rpc_meta_pb2.BuiltinService):
    def HeartBeat(self, rpc_controller, request, done):
        rsp = rpc_meta_pb2.HeartBeatResponse()
        if request.magic_num == 4321:
            rsp.return_code = 0
        else:
            rsp.return_code = 1

        return rsp


class RpcServer(object):
    def __init__(self, addr, service_timeout=10, spawn=1000):
        if isinstance(addr, str):
            self._addr = addr.split(':')  # addr string like '127.0.0.1:30006'
            self._addr[1] = int(self._addr[1])
        else:
            self._addr = addr

        if isinstance(spawn, (int, long)):
            self._pool = Pool(spawn)
            self._spawn = self._pool.spawn
        else:
            self._pool = None
            self._spawn = spawn

        self._services = {}
        self._service_timeout = service_timeout
        self._stat = RpcServerStat()
        self._stream_server = gevent.server.StreamServer(self._addr,
                                                         self._handle_connection)

        self._register_builtin_services()

    def _handle_connection(self, socket, addr):
        socket.setsockopt(gevent.socket.SOL_TCP, gevent.socket.TCP_NODELAY, 1)
        socket.setsockopt(gevent.socket.IPPROTO_TCP, gevent.socket.TCP_NODELAY, 1)

        rsp_queue = gevent.queue.Queue()
        is_connection_closed = [False]

        def call_service(req_info):
            meta_info, service, method, req = req_info
            self._stat.add_method_stat(meta_info.service_name,
                                       meta_info.method_name, 1)

            controller = RpcController()
            try:
                with Timeout(self._service_timeout):
                    rsp = service.CallMethod(method, controller, req, None)
            except Timeout:
                meta_info.has_error = True
                rsp = rpc_meta_pb2.ErrorResponse(err_code=rpc_meta_pb2.SERVER_SERVICE_TIMEOUT,
                                                 err_msg='service timeout')
            except:
                meta_info.has_error = True
                err_msg = 'Error calling service: ' + traceback.format_exc()
                rsp = rpc_meta_pb2.ErrorResponse(err_code=rpc_meta_pb2.SERVER_SERVICE_ERROR,
                                                 err_msg=err_msg)

            rsp_queue.put_nowait((meta_info, rsp))

        def recv_req():
            content = ""
            while True:
                try:
                    recv_buf = socket.recv(1024)
                    if len(recv_buf) == 0:
                        break
                except Exception, e:
                    logging.warning('recv_req error: ' + str(e))
                    break

                content += recv_buf
                mem_content = memoryview(content)
                cur_index = 0
                while cur_index < len(content):
                    if len(mem_content[cur_index:]) < 6:
                        break
                    elif mem_content[cur_index:cur_index + 2] != 'PB':
                        cur_index += 2  # skip the first 2 bytes
                        break

                    (buf_size,) = struct.unpack('!I',
                                                mem_content[cur_index + 2: cur_index + 6].tobytes())
                    if len(mem_content[cur_index + 6:]) < buf_size:
                        break

                    pb_buf = mem_content[cur_index + 6: cur_index + 6 + buf_size].tobytes()
                    cur_index += buf_size + 6
                    result = self.parse_message(pb_buf)
                    if result is None:
                        logging.warning('pb decode error, skip this message')
                        break

                    self._spawn(call_service, result)

                if cur_index > 0:
                    content = content[cur_index:]

            logging.info(str(addr) + 'has disconnected')
            is_connection_closed[0] = True

        def send_rsp():
            while not is_connection_closed[0]:
                try:
                    meta_info, rsp = rsp_queue.get(timeout=1)
                except gevent.queue.Empty:
                    continue

                serialized_rsp = serialize_message(meta_info, rsp)
                sent_bytes = 0
                try:
                    while sent_bytes < len(serialized_rsp):
                        sent_bytes += socket.send(serialized_rsp[sent_bytes:])
                except soc_error as e:
                    logging.warning('socket send error: ' + str(e))
                    break

        workers = [gevent.spawn(recv_req), gevent.spawn(send_rsp)]
        gevent.joinall(workers)

    def parse_message(self, buf):
        result = parse_meta(buf)
        if result is None:
            return None
        meta_len, pb_msg_len, meta_info = result

        # try to find the service
        try:
            service = self._services[meta_info.service_name]
        except KeyError:
            logging.warning('cannot find the service: ' + meta_info.service_name)
            return None

        method = service.GetDescriptor().FindMethodByName(meta_info.method_name)
        if method is None:
            logging.warning('cannot find the method: ' + meta_info.method_name)
            return None

        msg = parse_message(buf[8 + meta_len:8 + meta_len + pb_msg_len],
                             service.GetRequestClass(method))
        if msg is None:
            return None
        else:
            return meta_info, service, method, msg

    def register_service(self, service):
        self._services[service.GetDescriptor().full_name] = service

    def _register_builtin_services(self):
        self.register_service(BuiltinServiceImpl())

    def print_stat(self, interval):
        while True:
            gevent.sleep(interval)
            self._stat.print_stat()
            self._stat.reset_stat_per_min()

    def run(self, print_stat_interval=None):
        if print_stat_interval is not None and print_stat_interval > 0:
            stat_worker = gevent.spawn(self.print_stat, print_stat_interval)

        try:
            self._stream_server.serve_forever()
        finally:
            stat_worker.kill()
            if self._pool is not None:
                self._pool.join()